@@ -6,16 +6,20 @@ from collections import namedtuple
 import torch
 import torch.nn.functional as F
 from torch import nn, einsum
+from torch.utils.data import Dataset, DataLoader
+from torch.nn.utils.rnn import pad_sequence
 
 from einops import rearrange, reduce
 
 from toolformer_pytorch.palm import PaLM
 from toolformer_pytorch.optimizer import get_optimizer
+from toolformer_pytorch.prompts import DEFAULT_PROMPT_INPUT_TAG
 
 from beartype import beartype
-from beartype.typing import Callable, Optional, Union
+from beartype.typing import Callable, Optional, Union, List, Tuple
 
 from tqdm import tqdm
+from x_clip.tokenizer import tokenizer
 
 # helpers
 
@@ -25,6 +29,23 @@ def exists(val):
 def default(val, d):
     return val if exists(val) else d
 
+def identity(t):
+    return t
+
+def always(val):
+    def inner(*args, **kwargs):
+        return val
+    return inner
+
+def try_except(fn, callback = identity):
+    @wraps(fn)
+    def inner(*args):
+        try:
+            return fn(*args)
+        except Exception as e:
+            return callback(e)
+    return inner
+
 # tensor helpers
 
 def log(t, eps = 1e-20):
@@ -112,10 +133,7 @@ def replace_fn(
 
     # just return original text if there is some error with the function
 
-    try:
-        out = fn(*params)
-    except:
-        return orig_text
+    out = try_except(fn, always(None))(*params)
 
     # the api calling function can also arrest the process, by returning None
 
@@ -124,7 +142,7 @@ def replace_fn(
 
     # return original text with the output delimiter and the stringified output
 
-    return f'{text_without_end_api_token} {delimiter} {str(out)}{end_api_token}'
+    return f'{text_without_end_api_token} {delimiter} {str(out)} {end_api_token}'
 
 # main function, which takes a registry of functions, the text in question, and makes all the appropriate api calls and append the output
 
@@ -135,6 +153,9 @@ def create_function_regex(
     api_start_regex, api_stop_regex = map(re.escape, (api_start, api_stop))
     return rf'({api_start_regex}(\w+)\(([^)]*)\))({api_stop_regex})'
 
+def num_matches(substr: str, text: str):
+    return len(re.findall(re.escape(substr), text))
+
 def has_api_calls(
     text,
     api_start = ' [',
@@ -214,6 +235,7 @@ def sample(
     select_api_start_id_top_k = 10,
 ):
     device = next(model.parameters()).device
+    positions = positions.clone()
     max_seq_len = seq_len + 1
 
     # validate
@@ -243,7 +265,7 @@ def sample(
     # lengthen the prime to the entire sequence length
 
     remain_iterations = seq_len - prime_length
-    output = F.pad(prime, (max_seq_len - prime_length, 0), value = 0.)
+    output = F.pad(prime, (0, max_seq_len - prime_length), value = 0.)
 
     batch_indices = torch.arange(batch_size, device = device)
     batch_indices = rearrange(batch_indices, 'b -> b 1')
@@ -318,7 +340,6 @@ def sample(
     # remove the last token in output (use as noop placeholder)
 
     output = output[:, :-1]
-
     return output
 
 @beartype
@@ -430,11 +451,16 @@ def filter_tokens_with_api_response(
     assert all_contains_id(tokens_with_api_response, api_start_token_id)
     assert all_contains_id(tokens_with_api_response, api_end_token_id)
 
+    # auto set devices
+
+    device = next(model.parameters()).device
+    tokens, tokens_without_api_response, tokens_with_api_response = map(lambda t: t.to(device), (tokens, tokens_without_api_response, tokens_with_api_response))
+
     # get all the logits
 
     with torch.no_grad():
         model.eval()
-        logits, logits_without_api_response, logits_with_api_response = map(model, (tokens, tokens_with_api_response, tokens_without_api_response))
+        logits, logits_without_api_response, logits_with_api_response = map(model, (tokens, tokens_without_api_response, tokens_with_api_response))
 
     # derive all predicted prob of the actual next token id in sequence
 
@@ -451,8 +477,10 @@ def filter_tokens_with_api_response(
 
     # deriving the weighting for the original passage is more tricky
     # would need to start counting up from <api> start token location
+    # this would also assume that the language model perfectly copied the passage over and that both token ids are aligned except for the inserted API call - but this can be done with the custom filtering functions eventually
 
     weight = weight_and_mask_fn(tokens_without_api_response[:, 1:], api_start_token_id) # shift to the left by one since <api> does not exist in the original sequence
+    weight = weight[:, :probs.shape[-1]]
 
     # get the loss L for all three types of sequences
 
@@ -492,6 +520,42 @@ def filter_tokens_with_api_response(
 
     return ret
 
+# datasets and dataloaders
+
+# for bootstrapping the initial datasets with api calls
+# as well as for the final finetuning
+
+@beartype
+class PromptDataset(Dataset):
+    def __init__(
+        self,
+        prompt: str,
+        prompt_input_tag: str,
+        data: List[str],
+        tokenizer_encode: Callable
+    ):
+        self.data = data
+        self.prompt = prompt
+        self.prompt_input_tag_regex = re.escape(prompt_input_tag)
+
+    def __len__(self):
+        return len(self.data)
+
+    def __getitem__(self, idx):
+        data_string = self.data[idx]
+        data_with_prompt = re.sub(self.prompt_input_tag_regex, data_string, self.prompt)
+        token_ids = tokenizer.encode(data_with_prompt)
+        return torch.tensor(token_ids).long(), torch.tensor(len(token_ids)).long()
+
+def prompt_collate_fn(data, padding_value = 0):
+    prompts, prompt_lengths = zip(*data)
+    prompts = pad_sequence(prompts, batch_first = True, padding_value = padding_value)
+    return prompts, torch.stack(prompt_lengths)
+
+def PromptDataloader(ds: Dataset, *args, padding_value = 0, **kwargs):
+    collate_fn = partial(prompt_collate_fn, padding_value = padding_value)
+    return DataLoader(ds, *args, collate_fn = collate_fn, **kwargs)
+
 # classes
 
 @beartype
@@ -502,15 +566,182 @@ class Toolformer(nn.Module):
         *,
         tool_id: str,
         tool: Callable,
+        api_start_str = ' [',
+        api_stop_str = ']',
+        api_response_delimiter = '→',
+        api_start_id = None,
+        api_stop_id = None,
         teach_tool_prompt: str,
+        filter_threshold = 1.,
+        pad_id = 0,
+        prompt_batch_size = 4,
+        model_seq_len = 2048,
+        tokenizer_encode: Callable = tokenizer.encode,
+        tokenizer_decode: Callable = tokenizer.decode,
+        prompt_input_tag: str = DEFAULT_PROMPT_INPUT_TAG,
         exclude_filters: dict[str, Callable[[str], bool]] = dict()
     ):
         super().__init__()
         self.model = model
+        self.model_seq_len = model_seq_len
+
+        self.teach_tool_prompt = teach_tool_prompt
+        self.prompt_batch_size = prompt_batch_size
+        self.prompt_input_tag = prompt_input_tag
+
+        self.tokenizer_encode = tokenizer_encode
+        self.tokenizer_decode = tokenizer_decode
+        self.tokenizer_encode_to_tensor = lambda s: torch.tensor(tokenizer_encode(s)).long()
+
+        self.filter_threshold = filter_threshold
+
+        self.api_start_str = api_start_str
+        self.api_stop_str = api_stop_str
+        self.api_response_delimiter = api_response_delimiter
+
+        if not exists(api_start_id):
+            api_start_id = tokenizer_encode(api_start_str)
+            assert len(api_start_id) == 1
+            api_start_id = api_start_id[0]
+
+        self.api_start_id = api_start_id
+
+        if not exists(api_stop_id):
+            api_stop_id = tokenizer_encode(api_stop_str)
+            assert len(api_stop_id) == 1
+            api_stop_id = api_stop_id[0]
+
+        self.api_stop_id = api_stop_id
+
+        self.pad_id = pad_id
+
         self.tool_id = tool_id
         self.tool = tool
+        self.registry = {tool_id: tool}
+
+        assert num_matches(prompt_input_tag, teach_tool_prompt) == 1, f'there must be exactly one prompt input tag `{prompt_input_tag}` in your prompt to encourage the language model to use the designated tool'
+
         self.teach_tool_prompt = teach_tool_prompt
         self.exclude_filters = exclude_filters
 
-    def forward(self):
-        raise NotImplementedError
+    def generate_data_with_api_calls(
+        self,
+        data: List[str],
+        temperature: float = 0.9
+    ) -> List[str]:
+
+        dataset = PromptDataset(
+            data = data,
+            prompt_input_tag = self.prompt_input_tag,
+            prompt = self.teach_tool_prompt,
+            tokenizer_encode = self.tokenizer_encode
+        )
+
+        dl = PromptDataloader(
+            dataset,
+            batch_size = self.prompt_batch_size
+        )
+
+        prompted_outputs = []
+
+        for prime, positions in dl:
+
+            sampled_outputs = sample(
+                model = self.model,
+                prime = prime,
+                positions = positions,
+                seq_len = self.model_seq_len,
+                pad_id = self.pad_id,
+                temperature = temperature
+            )
+
+            for sample_output, position in zip(sampled_outputs, positions):
+                start_position = position.item()
+
+                prompted_output = self.tokenizer_decode(sample_output[start_position:])
+                prompted_outputs.append(prompted_output)
+
+        return prompted_outputs
+
+    def filter_and_keep_only_first_api_call(
+        self,
+        data,
+        data_with_api_calls: List[str],
+        return_excluded = False
+    ):
+        included = []
+        excluded = []
+
+        api_start_stop_kwargs = dict(api_start = self.api_start_str, api_stop = self.api_stop_str)
+
+        has_api_calls_ = partial(has_api_calls, **api_start_stop_kwargs)
+        replace_all_but_first_ = partial(replace_all_but_first, **api_start_stop_kwargs)
+
+        for datum, data_with_api_call in zip(data, data_with_api_calls):
+            if has_api_calls_(data_with_api_call):
+                data_with_api_call = replace_all_but_first_(data_with_api_call)
+                included.append((datum, data_with_api_call))
+            else:
+                excluded.append((datum, data_with_api_call))
+
+        included = tuple(map(list, zip(*included)))
+
+        if not return_excluded:
+            return included
+
+        excluded = tuple(map(list, zip(*excluded)))
+        return included, excluded
+
+    def make_api_calls(
+        self,
+        filtered_data_with_api_calls: List[str]
+    ):
+        invoke_tools_ = partial(
+            invoke_tools,
+            self.registry,
+            api_start = self.api_start_str,
+            api_stop = self.api_stop_str, delimiter = self.api_response_delimiter
+        )
+
+        data_with_api_responses = []
+        for data in filtered_data_with_api_calls:
+            output = invoke_tools_(data)
+            data_with_api_responses.append(output)
+
+        return data_with_api_responses
+
+    def filter_by_api_responses(
+        self,
+        data: List[str],
+        data_with_api_calls: List[str],
+        data_with_api_responses: List[str]
+    ) -> FilteredResults:
+
+        to_token_ids = lambda l: pad_sequence([*map(self.tokenizer_encode_to_tensor, l)], batch_first = True, padding_value = self.pad_id)
+
+        tokens, tokens_without_api_response, tokens_with_api_response = map(to_token_ids, (data, data_with_api_calls, data_with_api_responses))
+
+        filtered_results = filter_tokens_with_api_response(
+            model = self.model,
+            tokens = tokens,
+            tokens_with_api_response = tokens_with_api_response,
+            tokens_without_api_response = tokens_without_api_response,
+            filter_threshold = self.filter_threshold,
+            api_start_token_id = self.api_start_id,
+            api_end_token_id = self.api_stop_id
+        )
+
+        return filtered_results
+
+    def forward(
+        self,
+        data: List[str]
+    ):
+        data_with_api_calls = self.generate_data_with_api_calls(data)
+        filtered_data_with_api_calls = self.filter_and_keep_only_first_api_call(data_with_api_calls)
+
+        assert len(filtered_data_with_api_calls) > 0, 'your model failed to follow instructions and make API calls. please try a better model or do some better prompt engineering'
+
+        data_with_responses = self.make_api_calls(filtered_data_with_api_calls)
+
+        return data_with_responses
